/***************************************************************************

  M.A.M.E.UI  -  Multiple Arcade Machine Emulator with User Interface
  Win32 Portions Copyright (C) 1997-2003 Michael Soderstrom and Chris Kirmse,
  Copyright (C) 2003-2007 Chris Kirmse and the MAME32/MAMEUI team.

  This file is part of MAMEUI, and may only be used, modified and
  distributed under the terms of the MAME license, in "readme.txt".
  By continuing to use, modify or distribute this file you indicate
  that you have read the license and understand and accept it fully.

 ***************************************************************************/

/****************************************************************************
 *      datafile.c
 *      History database engine
 *
 *      Token parsing by Neil Bradley
 *      Modifications and higher-level functions by John Butler
 ****************************************************************************/
// standard windows headers
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

// standard C headers
#include <assert.h>
#include <ctype.h>

// MAME/MAMEUI headers
#include "emu.h"
#include "osdcomm.h"
#include "datafile.h"
#include "mui_opts.h" // For MameUIGlobal()
#include "sound/samples.h"

/****************************************************************************
 *      token parsing constants
 ****************************************************************************/

#define CR      0x0d    /* '\n' and '\r' meanings are swapped in some */
#define LF      0x0a    /*     compilers (e.g., Mac compilers) */

enum
{
        TOKEN_COMMA,
        TOKEN_EQUALS,
        TOKEN_SYMBOL,
        TOKEN_LINEBREAK,
        TOKEN_INVALID=-1
};

#define MAX_TOKEN_LENGTH        256


/****************************************************************************
 *      datafile constants
 ****************************************************************************/
#define DATAFILE_TAG '$'

static const char *DATAFILE_TAG_KEY = "$info";
static const char *DATAFILE_TAG_BIO = "$bio";
static const char *DATAFILE_TAG_MAME = "$mame";
static const char *DATAFILE_TAG_DRIV = "$drv";

char *g_history_filename = NULL;
char *g_mameinfo_filename = NULL;


/****************************************************************************
 *      private data for parsing functions
 ****************************************************************************/
static emu_file *fp;                                       /* Our file pointer */
static long dwFilePos;                                          /* file position */
static UINT8 bToken[MAX_TOKEN_LENGTH];          /* Our current token */

/* an array of driver name/drivers array index sorted by driver name
   for fast look up by name */
typedef struct
{
    const char *name;
    int index;
} driver_data_type;
static driver_data_type *sorted_drivers = NULL;
static int num_games;


/**************************************************************************
 **************************************************************************
 *
 *              Parsing functions
 *
 **************************************************************************
 **************************************************************************/

/*
 * DriverDataCompareFunc -- compare function for GetGameNameIndex
 */
static int CLIB_DECL DriverDataCompareFunc(const void *arg1,const void *arg2)
{
    return strcmp( ((driver_data_type *)arg1)->name, ((driver_data_type *)arg2)->name );
}

/*
 * GetGameNameIndex -- given a driver name (in lowercase), return
 * its index in the main drivers[] array, or -1 if it's not found.
 */
static int GetGameNameIndex(const char *name)
{
    driver_data_type *driver_index_info;
	driver_data_type key;
	key.name = name;

	if (sorted_drivers == NULL)
	{
		/* initialize array of game names/indices */
		int i;

		sorted_drivers = (driver_data_type *)malloc(sizeof(driver_data_type) * num_games);
		for (i=0;i<num_games;i++)
		{
			sorted_drivers[i].name = driver_list::driver(i).name;
			sorted_drivers[i].index = i;
		}
		qsort(sorted_drivers,num_games,sizeof(driver_data_type),DriverDataCompareFunc);
	}

	/* uses our sorted array of driver names to get the index in log time */
	driver_index_info = (driver_data_type *)bsearch(&key,sorted_drivers,num_games,sizeof(driver_data_type),
								DriverDataCompareFunc);

	if (driver_index_info == NULL)
		return -1;

	return driver_index_info->index;

}


/****************************************************************************
 *      Create an array with sorted sourcedrivers for the function
 *      index_datafile_drivinfo to speed up the datafile access
 ****************************************************************************/

typedef struct
{
    const char *srcdriver;
    int index;
} srcdriver_data_type;
static srcdriver_data_type *sorted_srcdrivers = NULL;


static int SrcDriverDataCompareFunc(const void *arg1,const void *arg2)
{
	return strcmp( ((srcdriver_data_type *)arg1)->srcdriver, ((srcdriver_data_type *)arg2)->srcdriver );
}


static int GetSrcDriverIndex(const char *srcdriver)
{
	srcdriver_data_type *srcdriver_index_info;
	srcdriver_data_type key;
	key.srcdriver = srcdriver;

	if (sorted_srcdrivers == NULL)
	{
		/* initialize array of game names/indices */
		int i;
		num_games = driver_list::total();

		sorted_srcdrivers = (srcdriver_data_type *)malloc(sizeof(srcdriver_data_type) * num_games);
		for (i=0;i<num_games;i++)
		{
			sorted_srcdrivers[i].srcdriver = driver_list::driver(i).source_file+17;
			sorted_srcdrivers[i].index = i;
		}
		qsort(sorted_srcdrivers,num_games,sizeof(srcdriver_data_type),SrcDriverDataCompareFunc);
	}

	srcdriver_index_info = (srcdriver_data_type *)bsearch(&key,sorted_srcdrivers,num_games,sizeof(srcdriver_data_type),
								SrcDriverDataCompareFunc);

	if (srcdriver_index_info == NULL)
		return -1;

	return srcdriver_index_info->index;

}


/****************************************************************************
 *      GetNextToken - Pointer to the token string pointer
 *                                 Pointer to position within file
 *
 *      Returns token, or TOKEN_INVALID if at end of file
 ****************************************************************************/
static UINT32 GetNextToken(UINT8 **ppszTokenText, long *pdwPosition)
{
        UINT32 dwLength;                                                /* Length of symbol */
        long dwPos;                                                             /* Temporary position */
        UINT8 *pbTokenPtr = bToken;                             /* Point to the beginning */
        UINT8 bData;                                                    /* Temporary data byte */

        while (1)
        {
                bData = fp->getc();                                  /* Get next character */

                /* If we're at the end of the file, bail out */

                if (fp->eof())
                        return(TOKEN_INVALID);

                /* If it's not whitespace, then let's start eating characters */

                if (' ' != bData && '\t' != bData)
                {
                        /* Store away our file position (if given on input) */

                        if (pdwPosition)
                                *pdwPosition = dwFilePos;

                        /* If it's a separator, special case it */

                        if (',' == bData || '=' == bData)
                        {
                                *pbTokenPtr++ = bData;
                                *pbTokenPtr = '\0';
                                ++dwFilePos;

                                if (',' == bData)
                                        return(TOKEN_COMMA);
                                else
                                        return(TOKEN_EQUALS);
                        }

                        /* Otherwise, let's try for a symbol */

                        if (bData > ' ')
                        {
                                dwLength = 0;                   /* Assume we're 0 length to start with */

                                /* Loop until we've hit something we don't understand */

                                while (bData != ',' &&
                                                 bData != '=' &&
                                                 bData != ' ' &&
                                                 bData != '\t' &&
                                                 bData != '\n' &&
                                                 bData != '\r' &&
                                                 fp->eof() == 0)
                                {
                                        ++dwFilePos;
                                        *pbTokenPtr++ = bData;  /* Store our byte */
                                        ++dwLength;
                                        assert(dwLength < MAX_TOKEN_LENGTH);
                                        bData = fp->getc();
                                }

                                /* If it's not the end of the file, put the last received byte */
                                /* back. We don't want to touch the file position, though if */
                                /* we're past the end of the file. Otherwise, adjust it. */

                                if (0 == fp->eof())
                                {
                                        fp->ungetc(bData);
                                }

                                /* Null terminate the token */

                                *pbTokenPtr = '\0';

                                /* Connect up the */

                                if (ppszTokenText)
                                        *ppszTokenText = bToken;

                                return(TOKEN_SYMBOL);
                        }

                        /* Not a symbol. Let's see if it's a cr/cr, lf/lf, or cr/lf/cr/lf */
                        /* sequence */

                        if (LF == bData)
                        {
                                /* Unix style perhaps? */

                                bData = fp->getc();          /* Peek ahead */
                                fp->ungetc(bData);          /* Force a retrigger if subsequent LF's */

                                if (LF == bData)                /* Two LF's in a row - it's a UNIX hard CR */
                                {
                                        ++dwFilePos;
                                        *pbTokenPtr++ = bData;  /* A real linefeed */
                                        *pbTokenPtr = '\0';
                                        return(TOKEN_LINEBREAK);
                                }

                                /* Otherwise, fall through and keep parsing. */

                        }
                        else
                        if (CR == bData)                /* Carriage return? */
                        {
                                /* Figure out if it's Mac or MSDOS format */

                                ++dwFilePos;
                                bData = fp->getc();          /* Peek ahead */

                                /* We don't need to bother with EOF checking. It will be 0xff if */
                                /* it's the end of the file and will be caught by the outer loop. */

                                if (CR == bData)                /* Mac style hard return! */
                                {
                                        /* Do not advance the file pointer in case there are successive */
                                        /* CR/CR sequences */

                                        /* Stuff our character back upstream for successive CR's */

                                        fp->ungetc(bData);

                                        *pbTokenPtr++ = bData;  /* A real carriage return (hard) */
                                        *pbTokenPtr = '\0';
                                        return(TOKEN_LINEBREAK);
                                }
                                else
                                if (LF == bData)        /* MSDOS format! */
                                {
                                        ++dwFilePos;                    /* Our file position to reset to */
                                        dwPos = dwFilePos;              /* Here so we can reposition things */

                                        /* Look for a followup CR/LF */

                                        bData = fp->getc();  /* Get the next byte */

                                        if (CR == bData)        /* CR! Good! */
                                        {
                                                bData = fp->getc();  /* Get the next byte */

                                                /* We need to do this to pick up subsequent CR/LF sequences */

                                                fp->seek(dwPos, SEEK_SET);

                                                if (pdwPosition)
                                                        *pdwPosition = dwPos;

                                                if (LF == bData)        /* LF? Good! */
                                                {
                                                        *pbTokenPtr++ = '\r';
                                                        *pbTokenPtr++ = '\n';
                                                        *pbTokenPtr = '\0';

                                                        return(TOKEN_LINEBREAK);
                                                }
                                        }
                                        else
                                        {
                                                --dwFilePos;
                                                fp->ungetc(bData);  /* Put the character back. No good */
                                        }
                                }
                                else
                                {
                                        --dwFilePos;
                                        fp->ungetc(bData);  /* Put the character back. No good */
                                }

                                /* Otherwise, fall through and keep parsing */
                        }
                }

                ++dwFilePos;
        }
}


/****************************************************************************
 *      ParseClose - Closes the existing opened file (if any)
 ****************************************************************************/
static void ParseClose(void)
{
        /* If the file is open, time for fclose. */

        if (fp)
        {
			fp->close();
			global_free(fp);
        }

        fp = NULL;
}


/****************************************************************************
 *      ParseOpen - Open up file for reading
 ****************************************************************************/
static BOOL ParseOpen(const char *pszFilename)
{
        /* Open file up in binary mode */
		fp = global_alloc(emu_file("", OPEN_FLAG_READ));
		/* If this is NULL, return FALSE. We can't open it */
		if (fp == NULL)
		{
			return FALSE;
		}

		file_error err = fp->open(pszFilename);
		if (err != FILERR_NONE)
		{
			fp->close();
			global_free(fp);
			fp = NULL;
			return FALSE;
		}

        /* Otherwise, prepare! */
        dwFilePos = 0;
        return TRUE;
}


/****************************************************************************
 *      ParseSeek - Move the file position indicator
 ****************************************************************************/
static UINT8 ParseSeek(long offset, int whence)
{
        int result = fp->seek(offset, whence);

        if (0 == result)
        {
                dwFilePos = fp->tell();
        }
        return (UINT8)result;
}



/**************************************************************************
 **************************************************************************
 *
 *              Datafile functions
 *
 **************************************************************************
 **************************************************************************/


/**************************************************************************
 *      ci_strncmp - case insensitive character array compare
 *
 *      Returns zero if the first n characters of s1 and s2 are equal,
 *      ignoring case.
 **************************************************************************/
static int ci_strncmp (const char *s1, const char *s2, int n)
{
        int c1, c2;

        while (n)
        {
                if ((c1 = tolower (*s1)) != (c2 = tolower (*s2)))
                        return (c1 - c2);
                else if (!c1)
                        break;
                --n;

                s1++;
                s2++;
        }
        return 0;
}


/**************************************************************************
 *      index_datafile
 *      Create an index for the records in the currently open datafile.
 *
 *      Returns 0 on error, or the number of index entries created.
 **************************************************************************/
static int index_datafile (struct tDatafileIndex **_index)
{
        struct tDatafileIndex *idx;
        int count = 0;
        UINT32 token = TOKEN_SYMBOL;
		num_games = driver_list::total();
        /* rewind file */
        if (ParseSeek (0L, SEEK_SET)) return 0;

        /* allocate index */
        idx = *_index = (tDatafileIndex *)malloc (num_games * sizeof (struct tDatafileIndex));
        if (NULL == idx) return 0;

        /* loop through datafile */
        while ((count < (num_games - 1)) && TOKEN_INVALID != token)
        {
                long tell;
                UINT8 *s;

                token = GetNextToken (&s, &tell);
                if (TOKEN_SYMBOL != token) continue;

                /* DATAFILE_TAG_KEY identifies the driver */
                if (!ci_strncmp (DATAFILE_TAG_KEY, (char *)s, strlen (DATAFILE_TAG_KEY)))
                {
                        token = GetNextToken (&s, &tell);
                        if (TOKEN_EQUALS == token)
                        {
                                int done = 0;

                                token = GetNextToken (&s, &tell);
                                while (!done && TOKEN_SYMBOL == token)
                                {
									int game_index;
									UINT8 *p;
									for (p = s; *p; p++)
										*p = tolower(*p);

									game_index = GetGameNameIndex((char *)s);
									if (game_index >= 0)
									{
										idx->driver = &driver_list::driver(game_index);
										idx->offset = tell;
										idx++;
										count++;
										done = 1;
										break;
									}
									if (!done)
									{
										token = GetNextToken (&s, &tell);

										if (TOKEN_COMMA == token)
											token = GetNextToken (&s, &tell);
										else
											done = 1; /* end of key field */
									}
                                }
                        }
                }
        }

        /* mark end of index */
        idx->offset = 0L;
        idx->driver = 0;
        return count;
}

static int index_datafile_drivinfo (struct tDatafileIndex **_index)
{
	struct tDatafileIndex *idx;
	int count = 0;
	UINT32 token = TOKEN_SYMBOL;
		num_games = driver_list::total();
	/* rewind file */
	if (ParseSeek (0L, SEEK_SET)) return 0;

	/* allocate index */
	idx = *_index = (tDatafileIndex *)malloc (num_games * sizeof (struct tDatafileIndex));
	if (NULL == idx) return 0;

	/* loop through datafile */
	while ((count < (num_games - 1)) && TOKEN_INVALID != token)
	{
		long tell;
		UINT8 *s;

		token = GetNextToken (&s, &tell);
		if (TOKEN_SYMBOL != token) continue;

		/* DATAFILE_TAG_KEY identifies the driver */
		if (!ci_strncmp (DATAFILE_TAG_KEY, (char *)s, strlen (DATAFILE_TAG_KEY)))
		{
			token = GetNextToken (&s, &tell);
			if (TOKEN_EQUALS == token)
			{
				int done = 0;

				token = GetNextToken (&s, &tell);
				while (!done && TOKEN_SYMBOL == token)
				{
					int src_index;

					strlwr((char *)s);
					src_index = GetSrcDriverIndex((char *)s);
					if (src_index >= 0)
					{
						idx->driver = &driver_list::driver(src_index);
						idx->offset = tell;
						idx++;
						count++;
						done = 1;
						break;
					}

					if (!done)
					{
						token = GetNextToken (&s, &tell);

						if (TOKEN_COMMA == token)
							token = GetNextToken (&s, &tell);
						else
							done = 1; /* end of key field */
					}
				}
			}
		}
	}

	/* mark end of index */
	idx->offset = 0L;
	idx->driver = 0;
	return count;
}

/**************************************************************************
 *      load_datafile_text
 *
 *      Loads text field for a driver into the buffer specified. Specify the
 *      driver, a pointer to the buffer, the buffer size, the index created by
 *      index_datafile(), and the desired text field (e.g., DATAFILE_TAG_BIO).
 *
 *      Returns 0 if successful.
 **************************************************************************/
static int load_datafile_text (const game_driver *drv, char *buffer, int bufsize,
        struct tDatafileIndex *idx, const char *tag)
{
        int     offset = 0;
        int found = 0;
        UINT32  token = TOKEN_SYMBOL;
        UINT32  prev_token = TOKEN_SYMBOL;

        *buffer = '\0';

        /* find driver in datafile index */
        while (idx->driver)
        {

                if (idx->driver == drv) break;

                idx++;
        }
        if (idx->driver == 0) return 1; /* driver not found in index */

        /* seek to correct point in datafile */
        if (ParseSeek (idx->offset, SEEK_SET)) return 1;

        /* read text until buffer is full or end of entry is encountered */
        while (TOKEN_INVALID != token)
        {
                UINT8 *s;
                int len;
                long tell;

                token = GetNextToken (&s, &tell);
                if (TOKEN_INVALID == token) continue;

                if (found)
                {
                        /* end entry when a tag is encountered */
                        if (TOKEN_SYMBOL == token && DATAFILE_TAG == s[0] && TOKEN_LINEBREAK == prev_token) break;

                        prev_token = token;

                        /* translate platform-specific linebreaks to '\n' */
                        if (TOKEN_LINEBREAK == token)
							strcpy ((char *)s, "\n");

                        /* append a space to words */
                        if (TOKEN_LINEBREAK != token)
							strcat ((char *)s, " ");

                        /* remove extraneous space before commas */
                        if (TOKEN_COMMA == token)
                        {
                                --buffer;
                                --offset;
                                *buffer = '\0';
                        }

                        /* Get length of text to add to the buffer */
                        len = strlen ((char *)s);

                        /* Check for buffer overflow */
                        /* For some reason we can get a crash if we try */
                        /* to use the last 30 characters of the buffer  */
                        if ((bufsize - offset) - len <= 45)
                        {
                            strcpy ((char *)s, " ...[TRUNCATED]");
                            len = strlen((char *)s);
                            strcpy (buffer, (char *)s);
                            buffer += len;
                            offset += len;
                            break;
                        }

                        /* add this word to the buffer */
                        strcpy (buffer, (char *)s);
                        buffer += len;
                        offset += len;
                }
                else
                {
                        if (TOKEN_SYMBOL == token)
                        {
                                /* looking for requested tag */
                                if (!ci_strncmp (tag, (char *)s, strlen (tag)))
                                        found = 1;
                                else if (!ci_strncmp (DATAFILE_TAG_KEY, (char *)s, strlen (DATAFILE_TAG_KEY)))
                                        break; /* error: tag missing */
                        }
                }
        }
        return (!found);
}

static int load_drivfile_text (const game_driver *drv, char *buffer, int bufsize,
	struct tDatafileIndex *idx, const char *tag)
{
	int	offset = 0;
	int found = 0;
	UINT32	token = TOKEN_SYMBOL;
	UINT32 	prev_token = TOKEN_SYMBOL;

	*buffer = '\0';

	/* find driver in datafile index */
	while (idx->driver)
	{
		if (idx->driver->source_file == drv->source_file) break;
		idx++;
	}
	if (idx->driver == 0) return 1;	/* driver not found in index */

	/* seek to correct point in datafile */
	if (ParseSeek (idx->offset, SEEK_SET)) return 1;

	/* read text until buffer is full or end of entry is encountered */
	while (TOKEN_INVALID != token)
	{
		UINT8 *s;
		int len;
		long tell;

		token = GetNextToken (&s, &tell);
		if (TOKEN_INVALID == token) continue;

		if (found)
		{
			/* end entry when a tag is encountered */
			if (TOKEN_SYMBOL == token && DATAFILE_TAG == s[0] && TOKEN_LINEBREAK == prev_token) break;

			prev_token = token;

			/* translate platform-specific linebreaks to '\n' */
			if (TOKEN_LINEBREAK == token)
				strcpy ((char *)s, "\n");

			/* append a space to words */
			if (TOKEN_LINEBREAK != token)
				strcat ((char *)s, " ");

			/* remove extraneous space before commas */
			if (TOKEN_COMMA == token)
			{
				--buffer;
				--offset;
				*buffer = '\0';
			}

			/* add this word to the buffer */
			len = strlen ((char *)s);
 			if ((len + offset) >= bufsize) break;
			strcpy (buffer, (char *)s);
			buffer += len;
			offset += len;
		}
		else
		{
			if (TOKEN_SYMBOL == token)
			{
				/* looking for requested tag */
				if (!ci_strncmp (tag, (char *)s, strlen (tag)))
					found = 1;
				else if (!ci_strncmp (DATAFILE_TAG_KEY, (char *)s, strlen (DATAFILE_TAG_KEY)))
					break; /* error: tag missing */
			}
		}
	}
	return (!found);
}

/**************************************************************************
 *      load_driver_history
 *      Load history text for the specified driver into the specified buffer.
 *      Combines $bio field of HISTORY.DAT with $mame field of MAMEINFO.DAT.
 *
 *      Returns 0 if successful.
 *
 *      NOTE: For efficiency the indices are never freed (intentional leak).
 **************************************************************************/
int load_driver_history (const game_driver *drv, char *buffer, int bufsize)
{
        static struct tDatafileIndex *hist_idx = 0;
        int history = 0;
        int err = 0;

        *buffer = 0;


        if (!g_history_filename || !*g_history_filename)
                g_history_filename = mame_strdup("history.dat");

        /* try to open history datafile */
        if (ParseOpen (g_history_filename))
        {
                /* create index if necessary */
                if (hist_idx)
                        history = 1;
                else
                        history = (index_datafile (&hist_idx) != 0);

                /* load history text */
                if (hist_idx)
                {
                        const game_driver *gdrv;

                        gdrv = drv;
                        do
                        {
//                                if ( ( gdrv->flags & GAME_IS_BIOS_ROOT) == 1 )
//					break;
                                err = load_datafile_text (gdrv, buffer, bufsize,
                                                                                  hist_idx, DATAFILE_TAG_BIO);
						int g = driver_list::clone(*gdrv);
						if (g!=-1) gdrv = &driver_list::driver(g); else gdrv = NULL;
                        } while (err && gdrv);

                        if (err) history = 0;
                }
                ParseClose ();
        }

	strcat(buffer, "\n");
	return (history == 0);
}

int load_driver_mameinfo (const game_driver *drv, char *buffer, int bufsize)
{
	static struct tDatafileIndex *mame_idx = 0;
	machine_config config(*drv,MameUIGlobal());
	const game_driver *parent = NULL;

	char name[300];
	int mameinfo = 0;
	int err = 0;
	int i;
	int has_sound;
	int is_bios = 0;

	*buffer = 0;

	strcat(buffer, "\nMAMEINFO:\n");

	/* List the game info 'flags' */
	if (drv->flags & GAME_NOT_WORKING)
		strcat(buffer, "THIS GAME DOESN'T WORK PROPERLY\n");
	if (drv->flags & GAME_UNEMULATED_PROTECTION)
		strcat(buffer, "The game has protection which isn't fully emulated.\n");
	if (drv->flags & GAME_IMPERFECT_GRAPHICS)
		strcat(buffer, "The video emulation isn't 100% accurate.\n");
	if (drv->flags & GAME_WRONG_COLORS)
		strcat(buffer, "The colors are completely wrong.\n");
	if (drv->flags & GAME_IMPERFECT_COLORS)
		strcat(buffer, "The colors aren't 100% accurate.\n");
	if (drv->flags & GAME_NO_SOUND)
		strcat(buffer, "The game lacks sound.\n");
	if (drv->flags & GAME_IMPERFECT_SOUND)
		strcat(buffer, "The sound emulation isn't 100% accurate.\n");
	if (drv->flags & GAME_SUPPORTS_SAVE)
		strcat(buffer, "Save state support.\n");
	if (drv->flags & GAME_NO_COCKTAIL)
		strcat(buffer, "Screen flipping in cocktail mode is not supported.\n");
	if (drv->flags & GAME_MECHANICAL)
		strcat(buffer, "The game contains mechanical parts.\n");
		strcat(buffer, "\n");
	if (drv->flags & GAME_IS_BIOS_ROOT)
		is_bios = 1;

	if (!g_mameinfo_filename || !*g_mameinfo_filename)
		g_mameinfo_filename = mame_strdup("mameinfo.dat");

	/* try to open mameinfo datafile */
	if (ParseOpen (g_mameinfo_filename))
	{
		/* create index if necessary */
		if (mame_idx)
			mameinfo = 1;
		else
			mameinfo = (index_datafile (&mame_idx) != 0);

		/* load informational text (append) */
		if (mame_idx)
		{
			int len = strlen (buffer);
			const game_driver *gdrv;

			gdrv = drv;
			do
			{
				err = load_datafile_text (gdrv, buffer+len, bufsize-len, mame_idx, DATAFILE_TAG_MAME);

				int g = driver_list::clone(*gdrv);
				if (g!=-1) gdrv = &driver_list::driver(g); else gdrv = NULL;

			} while (err && gdrv);

			if (err) mameinfo = 0;
		}
		ParseClose ();
	}


	/* GAME INFORMATIONS */

	sprintf(name,"\nGAME: %s\n", drv->name);
	strcat(buffer, name);
	sprintf(name,"%s", drv->description);
	strcat(buffer, name);
	sprintf(name," (%s %s)\n\nCPU:\n", drv->manufacturer, drv->year);
	strcat(buffer, name);

	execute_interface_iterator iter(config.root_device());
	device_execute_interface *cpu = iter.first();
	while (cpu)
	{
		if (cpu->device().clock()-((cpu->device().clock()/1000000)*1000000) > 0)
			sprintf(name,"%s (%d Hz)\n", cpu->device().name(), cpu->device().clock());
		else
			sprintf(name,"%s (%d MHz)\n", cpu->device().name(), cpu->device().clock()/1000000);

		strcat(buffer, name);

		cpu = iter.next();
	}


	strcat(buffer, "\nSOUND:\n");
	i = 0; has_sound = 0;
	
	/* iterate over sound chips */
	sound_interface_iterator sounditer(config.root_device());
	const device_sound_interface *sound = sounditer.first();
	while(sound)
	{
		int clock, count;
		device_type sound_type_;
		char tmpname[1024];

		sprintf(tmpname,"%s",sound->device().name());

		sound_type_ = sound->device().type();
		clock = sound->device().clock();

		has_sound = 1;
		count = 1;
		sound = sounditer.next();
		/* Matching chips at the same clock are aggregated */
		while (sound
			&& sound->device().type() == sound_type_
			&& sound->device().clock() == clock)
		{
			count++;
			sound = sounditer.next();
		}

		if (count > 1)
		{
			sprintf(name,"%dx ",count);
			strcat(buffer, name);
		}
		strcat(buffer, tmpname);

		if (clock)
		{
			if (clock-((clock/1000000)*1000000) > 0)
				sprintf(name," (%d Hz)", clock);
			else
				sprintf(name," (%d MHz)", clock/1000000);
			strcat(buffer, name);
		}

		strcat(buffer,"\n");
	}

	if (has_sound)
	{
		speaker_device_iterator iter(config.root_device());
		int channels = iter.count();
		if(channels == 1)
			sprintf(name,"%d Channel\n",channels);
		else
			sprintf(name,"%dx Channels\n",channels);
		strcat(buffer, name);
	}



	strcat(buffer, "\nVIDEO:\n");
	screen_device_iterator screeniter(config.root_device());
	const screen_device *screen = screeniter.first();
	if (screen == NULL)
		strcat(buffer, "None\n");
	else
	{
		for (; screen != NULL; screen = screeniter.next())
		{
			if (drv->flags & ORIENTATION_SWAP_XY)
				sprintf(name,"%dx%d (V)", screen->visible_area().height(), screen->visible_area().width());
			else
				sprintf(name,"%dx%d (H)", screen->visible_area().width(), screen->visible_area().height());

			strcat(buffer, name);

			i = ATTOSECONDS_TO_HZ(screen->refresh_attoseconds());
			if ((ATTOSECONDS_TO_HZ(screen->refresh_attoseconds()) * 1000000 - i * 1000000) > 0)
				sprintf(name," (%f Hz, ", ATTOSECONDS_TO_HZ(screen->refresh_attoseconds()));
			else
				sprintf(name," (%d Hz, ", i);

			strcat(buffer, name);

			if (screen->screen_type() == SCREEN_TYPE_VECTOR)
				sprintf(name,"palette %d, vector)\n", config.m_total_colors);
			else
				sprintf(name,"palette %d, raster)\n", config.m_total_colors);
			strcat(buffer, name);
		}
	}



	strcat(buffer, "\nROM REGION:\n");

	int g = driver_list::clone(*drv);
	if (g!=-1) parent = &driver_list::driver(g);

	device_iterator deviter(config.root_device());
	for (device_t *device = deviter.first(); device != NULL; device = deviter.next())
		for (const rom_entry *region = rom_first_region(*device); region != NULL; region = rom_next_region(region))
			for (const rom_entry *rom = rom_first_file(region); rom != NULL; rom = rom_next_file(rom))
			{
				hash_collection hashes(ROM_GETHASHDATA(rom));
				int in_parent = 0;

				if (g!=-1)
				{
					machine_config pconfig(*parent, MameUIGlobal());
					device_iterator deviter(pconfig.root_device());
					for (device_t *device = deviter.first(); device != NULL; device = deviter.next())
						for (const rom_entry *pregion = rom_first_region(*device); pregion != NULL; pregion = rom_next_region(pregion))
							for (const rom_entry *prom = rom_first_file(pregion); prom != NULL; prom = rom_next_file(prom))
							{
								hash_collection phashes(ROM_GETHASHDATA(prom));
								if (hashes == phashes)
								{
									in_parent = 1;
									break;
								}
							}
				}

				sprintf(name,"%-12s ",ROM_GETNAME(rom));
				strcat(buffer, name);
				sprintf(name,"%6x ", rom_file_size(rom));
				strcat(buffer, name);
				strcat(buffer, ROMREGION_GETTAG(region));
				sprintf(name," %7x",ROM_GETOFFSET(rom));
				strcat(buffer, name);

				if (!ROMREGION_ISDISKDATA(region) && in_parent)
					strcat(buffer, "  merge");
				if (hashes.flag(hash_collection::FLAG_NO_DUMP))
					strcat(buffer, "  NO_DUMP");
				if (hashes.flag(hash_collection::FLAG_BAD_DUMP))
					strcat(buffer, "  BAD_DUMP");
				strcat(buffer, "\n");
			}


#if 0
  strcat(buffer, "\nINPUT:\n");
  if (drv->ipt != NULL)
  {
		input_port_config *port;
		ioport_list portlist;
		astring errors;
		device_iterator iter(config.root_device());
		for (device_t *cfg = iter.first(); cfg != NULL; cfg = iter.next())
		{
			if (cfg->input_ports()!=NULL) {
				input_port_list_init(*cfg, portlist, errors);
			}
		}


	enum
	{
		ANALOG_TYPE_JOYSTICK,
		ANALOG_TYPE_DIAL,
		ANALOG_TYPE_TRACKBALL,
		ANALOG_TYPE_PADDLE,
		ANALOG_TYPE_LIGHTGUN,
		ANALOG_TYPE_PEDAL,
		ANALOG_TYPE_COUNT
	};

	const UINT8 DIR_LEFTRIGHT = 0x01;
	const UINT8 DIR_UPDOWN = 0x02;
	const UINT8 DIR_4WAY = 0x04;
	const UINT8 DIR_DUAL = 0x08;


	struct
	{
		const char *	type;			/* general type of input */
		bool			analog;
		bool			keyb;
		INT32			min;			/* analog minimum value */
		INT32			max;			/* analog maximum value  */
		INT32			sensitivity;	/* default analog sensitivity */
		INT32			keydelta;		/* default analog keydelta */
		bool			reverse;		/* default analog reverse setting */
	} control_info[ANALOG_TYPE_COUNT];


	memset(&control_info, 0, sizeof(control_info));

	int nplayer = 0;
	int nbutton = 0;
	int ncoin = 0;
	UINT8 joytype = 0;
	bool service = false;
	bool tilt = false;
	bool keypad = false;
	bool keyboard = false;


	for (input_port_config *port = portlist.first(); port != NULL; port = port->next())
		for (input_field_config *field = port->fieldlist().first(); field != NULL; field = field->next())
		{
			int analogtype = -1;

			if (nplayer < field->player + 1)
				nplayer = field->player + 1;

			switch (field->type)
			{
				case IPT_JOYSTICKRIGHT_LEFT:
				case IPT_JOYSTICKRIGHT_RIGHT:
				case IPT_JOYSTICKLEFT_LEFT:
				case IPT_JOYSTICKLEFT_RIGHT:
					joytype |= DIR_DUAL;

				case IPT_JOYSTICK_LEFT:
				case IPT_JOYSTICK_RIGHT:
					joytype |= DIR_LEFTRIGHT | ((field->way == 4) ? DIR_4WAY : 0);
					break;

				case IPT_JOYSTICKRIGHT_UP:
				case IPT_JOYSTICKRIGHT_DOWN:
				case IPT_JOYSTICKLEFT_UP:
				case IPT_JOYSTICKLEFT_DOWN:
					joytype |= DIR_DUAL;

				case IPT_JOYSTICK_UP:
				case IPT_JOYSTICK_DOWN:
					joytype |= DIR_UPDOWN | ((field->way == 4) ? DIR_4WAY : 0);
					break;

				case IPT_PADDLE:
					control_info[analogtype = ANALOG_TYPE_PADDLE].type = "Paddle";
					break;

				case IPT_DIAL:
					control_info[analogtype = ANALOG_TYPE_DIAL].type = "Dial";
					analogtype = ANALOG_TYPE_DIAL;
					break;

				case IPT_TRACKBALL_X:
				case IPT_TRACKBALL_Y:
					control_info[analogtype = ANALOG_TYPE_TRACKBALL].type = "Trackball";
					analogtype = ANALOG_TYPE_TRACKBALL;
					break;

				case IPT_AD_STICK_X:
				case IPT_AD_STICK_Y:
					control_info[analogtype = ANALOG_TYPE_JOYSTICK].type = "Stick";
					break;

				case IPT_LIGHTGUN_X:
				case IPT_LIGHTGUN_Y:
					control_info[analogtype = ANALOG_TYPE_LIGHTGUN].type = "Lightgun";
					break;

				case IPT_PEDAL:
				case IPT_PEDAL2:
				case IPT_PEDAL3:
					control_info[analogtype = ANALOG_TYPE_PEDAL].type = "Pedal";
					break;

				case IPT_BUTTON1:
				case IPT_BUTTON2:
				case IPT_BUTTON3:
				case IPT_BUTTON4:
				case IPT_BUTTON5:
				case IPT_BUTTON6:
				case IPT_BUTTON7:
				case IPT_BUTTON8:
				case IPT_BUTTON9:
				case IPT_BUTTON10:
				case IPT_BUTTON11:
				case IPT_BUTTON12:
				case IPT_BUTTON13:
				case IPT_BUTTON14:
				case IPT_BUTTON15:
				case IPT_BUTTON16:
					nbutton = MAX(nbutton, field->type - IPT_BUTTON1 + 1);
					break;

				case IPT_COIN1:
				case IPT_COIN2:
				case IPT_COIN3:
				case IPT_COIN4:
				case IPT_COIN5:
				case IPT_COIN6:
				case IPT_COIN7:
				case IPT_COIN8:
					ncoin = MAX(ncoin, field->type - IPT_COIN1 + 1);
					break;

				case IPT_KEYPAD:
					keypad = true;
					break;

				case IPT_KEYBOARD:
					keyboard = true;
					break;

				case IPT_SERVICE:
					service = true;
					break;

				case IPT_TILT:
					tilt = true;
					break;
			}

			if (analogtype != -1)
			{
				if (field->min != 0)
					control_info[analogtype].min = field->min;
				if (field->max != 0)
					control_info[analogtype].max = field->max;
				if (field->sensitivity != 0)
					control_info[analogtype].sensitivity = field->sensitivity;
				if (field->delta != 0)
					control_info[analogtype].keydelta = field->delta;
				if ((field->flags & ANALOG_FLAG_REVERSE) != 0)
					control_info[analogtype].reverse = true;
			}
		}


	if (nplayer == 1) {sprintf(name,"%d Player", nplayer); strcat(buffer, name);}
	if (nplayer > 1) {sprintf(name,"%d Players", nplayer); strcat(buffer, name);}
	if (nbutton == 1) {sprintf(name,", %d Button", nbutton); strcat(buffer, name);}
	if (nbutton > 1) {sprintf(name,", %d Buttons", nbutton); strcat(buffer, name);}
	if (ncoin == 1) {sprintf(name,", %d Coin Slot", ncoin);	strcat(buffer, name);}
	if (ncoin > 1) {sprintf(name,", %d Coin Slots", ncoin); strcat(buffer, name);}
	if (service) strcat(buffer, ", Service");
	if (tilt) strcat(buffer, ", Tilt");

	strcat(buffer, "\n");

	if (joytype != 0)
	{
		const char *vertical = ((joytype & DIR_LEFTRIGHT) == 0) ? "(vertical)" : "";
		const char *doubletype = ((joytype & DIR_DUAL) != 0) ? "Double " : "Joystick ";
		const char *way = ((joytype & DIR_LEFTRIGHT) == 0 || (joytype & DIR_UPDOWN) == 0) ? "2-way " : ((joytype & DIR_4WAY) != 0) ? "4-way " : "8-way ";
		sprintf(name,"%s%s%s\n", way, doubletype, vertical);
		strcat(buffer, name);
	}

	for (int type = 0; type < ANALOG_TYPE_COUNT; type++)
		if (control_info[type].type != NULL)
		{
			sprintf(name, "%s  (", control_info[type].type);
			strcat(buffer, name);
			if (control_info[type].min != 0 || control_info[type].max != 0)
			{
				sprintf(name, "min=%d", control_info[type].min);
				strcat(buffer, name);
				sprintf(name, " max=%d", control_info[type].max);
				strcat(buffer, name);
			}
			if (control_info[type].sensitivity != 0)
			{
				sprintf(name, " sensitiv=%d", control_info[type].sensitivity);
				strcat(buffer, name);
			}
			if (control_info[type].keydelta != 0)
			{
				sprintf(name, " keydelta=%d", control_info[type].keydelta);
				strcat(buffer, name);
			}
			if (control_info[type].reverse)
				strcat(buffer, " reverse");

			strcat(buffer, ")\n");
		}

	if (keypad)
		strcat(buffer, "Keypad\n");
	if (keyboard)
		strcat(buffer, "Keyboard\n");



	strcat(buffer, "\nDIPSWITCHES:");
	for (port = portlist.first(); port != NULL; port = port->next())
	{
		input_field_config *field;
		for (field = port->fieldlist().first(); field != NULL; field = field->next())
			if (field->type == IPT_DIPSWITCH)
			{
				input_setting_config *setting;

				sprintf(name,"\n%s: ", input_field_name(field));
				strcat(buffer, name);
				for (setting = field->settinglist().first(); setting != NULL; setting = setting->next())
				{
					if (setting->value == field->defvalue)
						sprintf(name,"[%s] ", setting->name);
					else
						sprintf(name,"%s ", setting->name);

					strcat(buffer, name);
				}
			}
	}

	strcat(buffer, "\n");

  }
#endif

	
	samples_device_iterator samplesiter(config.root_device());
	for (samples_device *device = samplesiter.first(); device != NULL; device = samplesiter.next())
	{
		samples_iterator sampiter(*device);
		if (sampiter.altbasename() != NULL)
		{
			sprintf(name,"\nSAMPLES (%s):\n", sampiter.altbasename());
			strcat(buffer, name);
		}

		tagmap_t<int> already_printed;
		for (const char *samplename = sampiter.first(); samplename != NULL; samplename = sampiter.next())
		{
			// filter out duplicates
			if (already_printed.add(samplename, 1) == TMERR_DUPLICATE)
				continue;

			// output the sample name
			sprintf(name,"%s.wav\n", samplename);
			strcat(buffer, name);
		}
	}

	if (!is_bios)
	{
		int g = driver_list::clone(*drv);
		if (g!=-1) drv = &driver_list::driver(g);

		strcat(buffer, "\nORIGINAL:\n");
		strcat(buffer, drv->description);
		strcat(buffer, "\n\nCLONES:\n");
		for (i = 0; i < driver_list::total(); i++)
		{
			if (!strcmp (drv->name, driver_list::driver(i).parent)) 
			{
				strcat(buffer, driver_list::driver(i).description);
				strcat(buffer, "\n");
			}
		}
	}



	strcat(buffer, "\n");
	return (mameinfo == 0);
}



int load_driver_drivinfo (const game_driver *drv, char *buffer, int bufsize)
{
	static struct tDatafileIndex *driv_idx = 0;
	int drivinfo = 0;
	int err;
	int i;

	*buffer = 0;

	/* Print source code file */
	sprintf (buffer, "\nSOURCE: %s\n", drv->source_file+17);

	if (!g_mameinfo_filename || !*g_mameinfo_filename)
		g_mameinfo_filename = mame_strdup("mameinfo.dat");

	/* Try to open mameinfo datafile - driver section*/
	if (ParseOpen (g_mameinfo_filename))
	{
		/* create index if necessary */
		if (driv_idx)
			drivinfo = 1;
		else
			drivinfo = (index_datafile_drivinfo (&driv_idx) != 0);

		/* load informational text (append) */
		if (driv_idx)
		{
			int len = strlen (buffer);
			err = load_drivfile_text (drv, buffer+len, bufsize-len, driv_idx, DATAFILE_TAG_DRIV);
			if (err) drivinfo = 0;
		}
		ParseClose ();
	}

	strcat(buffer,"\nGAMES SUPPORTED:\n");
	for (i = 0; i < driver_list::total(); i++)
	{

		if (!strcmp (drv->source_file+17, driver_list::driver(i).source_file+17) && !(driver_list::driver(i).flags & GAME_IS_BIOS_ROOT)) 
		{
			strcat(buffer, driver_list::driver(i).description);
			strcat(buffer,"\n");
		}
	}


	return (drivinfo == 0);

}
